/*
        PURPOSE -- Data Conversion MACROS for the ifs library, MC-68020
 
        RESTRICTIONS --
		MC-68020 version = assumes "machine data" in the 68020 format
		Controlled by #define M2E (defined = Machine To External),
	       	defaulting to E2M (External to Machine)
 
	Macro invocation:
 
	CVTxxx_yyy (ibyte, obyte, nbyte )				
 
	xxx   - External data type: ( EEE, VAX, 386, BYT )	
	yyy   - Element type: ( SHRT, LONG, REAL, DBLE )
 
	ibyte - input  buffer char* pointer, not incremented.
	obyte - output buffer char* pointer, not incremented.
	nbyte - input bytes   long int byte count
 
	Byte ordering assumptions:
	Bit number:                    31 24 23 16 15 08 07 00
	VAX byte order:                 n+3   n+2   n+1   n+0
	68020 byte order:               n+0   n+1   n+2   n+3
	80386 byte order:               n+3   n+2   n+1   n+0
	(bytes assumed to use the same most significant bit, this
	 will NOT be the case for IBM-360 data)
 
	Note:	Some of the macros are bi-directional, using the same
		byte swap algorithm for both M->E and E->M.
 
	Note:	The SEDEEE means MC-68020 IEEE data = straight dump
		The SED386 means I-80386 IEEE external data (byte swap)
c		The SEDBYT means "byte" = no conversion
		The SEDVAX menas VAX data = complex conversions
*/
/*
	INTeger	VAX to 68020 IEEE to VAX conversions
	Only byte swap is necessary.
	For the VAX, this code is bi-directional (same for M2E and E2M)
*/
#define CVTEEE_SHORT( ibyte, obyte, nbyte )				\
	{ memcpy (obyte,ibyte,nbyte); }
#define CVT386_SHORT( ibyte, obyte, nbyte ) 				\
	{ short *ip, *op, *last = (short*) (ibyte + nbyte);		\
	  op  = ( short *) obyte;					\
	  ip  = ( short *) ibyte;					\
	  while ( ip < last )						\
	   { *op++ = ifs_swap_half(*ip); ip++; };			\
	}
#define CVTVAX_SHORT( ibyte, obyte, nbyte ) 				\
	{ short *ip, *op, *last = (short*) (ibyte + nbyte);		\
	  op  = ( short *) obyte;					\
	  ip  = ( short *) ibyte;					\
	  while ( ip < last )						\
	   { *op++ = ifs_swap_half(*ip); ip++; };			\
	}
#define CVTBYT_SHORT( ibyte, obyte, nbyte ) 				\
	{ memcpy (obyte,ibyte,nbyte); }
 
#define CVTEEE_LONG( ibyte, obyte, nbyte )				\
	{ memcpy (obyte,ibyte,nbyte); }
#define CVT386_LONG( ibyte, obyte, nbyte ) 				\
	{ long *ip, *op, *last = (long*) (ibyte + nbyte);		\
	  op  = ( long *) obyte;					\
	  ip  = ( long *) ibyte;					\
	  while ( ip < last )						\
	   { *op++ = ifs_swap_word(*ip); ip++; };			\
	}
#define CVTVAX_LONG( ibyte, obyte, nbyte ) 				\
	{ long *ip, *op, *last = (long*) (ibyte + nbyte);		\
	  op  = ( long *) obyte;					\
	  ip  = ( long *) ibyte;					\
	  while ( ip < last )						\
	   { *op++ = ifs_swap_word(*ip); ip++; };			\
	}
#define CVTBYT_LONG( ibyte, obyte, nbyte ) 				\
	{ memcpy (obyte,ibyte,nbyte); }
 
/*
	R E A L  (FLOAT)  FORMAT  CONVERSIONS
 
	IEEE Float format, normalized, "hidden" highest order mantissa bit,
	2 based exponent biassed +127:
                   IEEE	 		    
        ------------------------------      
	|S|expone|      mantissa     |	    
        ------------------------------      
       31 30   23 22                0       
       (1)  (8)          (23)       
 
	DEC VAX Double format, normalized, "hidden" highest order mantissa bit,
	2 based exponent biased +129 (128+1 for a hidden bit)
                  DEC VAX
        ------------------------------
	|  mantissa   |S|expone|manti|   
        ------------------------------
	31          16   14   7 6    0    
	   (low 16)        (8)  (hi 7)
	
	For DEC->IEEE: subtract DEC  base(129), add IEEE base(127) => exp -=2
	For IEEE->DEC: subtract IEEE base(127), add DEC  base(129) => exp +=2
	On potential over/under flow, we force max.number resp. zero.
	In both direcions we assume user's buffer is properly alligned.
*/
#define CVTEEE_REAL( ibyte, obyte, nbyte )				\
	{ memcpy (obyte,ibyte,nbyte); }
#define CVT386_REAL( ibyte, obyte, nbyte )				\
	{  long *last, *ip, *op;				      \
	last = ( long *) (ibyte + nbyte);			      \
	ip = ( long *) ibyte;					      \
	op = ( long *) obyte;					      \
	while (ip < last ) {                                          \
	    *op++ = ifs_swap_word(*ip);                               \
	    ip++;                                                     \
	    };                                                        \
	}
#ifdef  M2E				
/*
	FLOAT 68020 IEEE to VAX conversion
	Float IEEE is copied into local "eee0" longword store
	Exponent is adjusted to VAX base limiting underflow.
	Bytes are outputted in the VAX float words/bytes order.
*/
#define CVTVAX_REAL( ibyte, obyte, nbyte )				\
	{ long *last = (ibyte + nbyte); long *op, *vax0;		\
	  op = (long*) obyte; vax0 = (long*) ibyte; /* input FLOAT */	\
	  while ( vax0 < last )						\
	   { long exp, eee0;			    /* eee format    */ \
	     eee0 = *vax0;			    /* get input IEEE*/	\
	     if( eee0 )				    /* nonzero data  */ \
	      { exp  = eee0 & 0x7F800000; 	    /* separate exp  */ \
	        eee0 = eee0 & 0x807FFFFF;   	    /* remove expon. */ \
	        if(exp > 0x7E800000) 		    /* exp overflow? */	\
		  eee0 |= 0x7FFFEEEE;    	    /* = force max   */ \
	        else				    /*   (signed)    */ \
	          eee0 |= (exp + 0x01000000);	    /* exp +=2       */ \
	      }								\
	     if ( eee0 == 0x80000000 ) eee0 = 0;    /* neg ZERO?     */	\
	     *op++ = ifs_swap_bytes(eee0);	    /* SWAP to VAX   */ \
	     vax0++;				    /* next long     */ \
	   };  								\
	}
#else /*E2M*/
/*
	FLOAT VAX to 68020 IEEE machine conversion
	Operation performed in the output buffer (long alignement assumed)
	Bytes are swapped to build IEEE float bit order
	Exponent is adjusted to IEEE notation 
*/
#define CVTVAX_REAL( ibyte, obyte, nbyte ) 				\
	{ long *last = (long*) (ibyte + nbyte);	long *eee0, *ip;	\
	  eee0 = (long*) obyte; ip = (long*) ibyte; /* init. pointers*/ \
	  while ( ip < last )						\
	   { long  exp;							\
	     *eee0 = ifs_swap_bytes(*ip);		/* SWAP bytes*/ \
	     if ( *eee0 )			    /* nonzero data  */ \
	      { exp  = *eee0 & 0x7F800000; 	    /* extract expon.*/ \
	        if(exp  < 0x01000000) 		    /* exp underflow?*/	\
		  *eee0 = 0x00000000;		    /* = force zero  */ \
	        else				    /* blank exponent*/	\
	          *eee0 =(*eee0 & 0x807FFFFF) | (exp-0x1000000);/* exp-=2*/ \
	      }								\
	     else							\
		*eee0 = 0;						\
	    eee0++; ip++;			    /* next longs    */ \
	   };  								\
	}
#endif
 
#define CVTBYT_REAL( ibyte, obyte, nbyte ) 				\
	{ memcpy (obyte,ibyte,nbyte); }
/*
	D O U B L E   FORMAT CONVERSION
	===============================
 
	IEEE Double format, normalized, "hidden" highest order mantissa bit,
	exponent biassed +1023:
 
        --------------------------    -------------------------
	|S| expo |    mantissa   |    |       mantissa        |
        --------------------------    -------------------------
        31 30  20 19             0      31                    0
        (1) (11)         (20)                     (32)
 
	DEC VAX Double format, normalized, "hidden" highest order mantissa bit,
	exponent biased +129 (128+1 for a hidden bit)
 
        --------------------------    -------------------------
	| mantissa  |S|expo|manti|    | low-mant  | high-mant |
        --------------------------    -------------------------
	31        16   14 7 6   0      31       16 15        0
            (16)     1  (8)  (7)          (16)         (16)
 
	In both direcions we do not worry about alignment since machine data
	alignment is user's problem, and ext. data is byte accessed.
*/
#define CVTEEE_DBLE( ibyte, obyte, nbyte )				\
	{ memcpy (obyte,ibyte,nbyte); }
 
#define CVT386_DBLE( ibyte, obyte, nbyte )				\
	{ char *last = (ibyte + nbyte);					\
	  while ( ibyte < last )  					\
	  { *obyte++ = ibyte[3]; *obyte++ = ibyte[2];  /* 68020 / 386*/	\
	    *obyte++ = ibyte[1]; *obyte++ = ibyte[0];  /* only use   */	\
	    *obyte++ = ibyte[7]; *obyte++ = ibyte[6];  /* rev. order */	\
	    *obyte++ = ibyte[5]; *obyte++ = ibyte[4]; ibyte += 8;	\
	  }								\
	}
#ifdef  M2E
/*
	DOUBLE 68020 IEEE to VAX conversion
	Input is assumed alligned IEEE 68020 format
	Output is built in temporrary vax0, vax1 longwords:
	 Exponent changed to a new base (-1023+129); uderflow = zero,
	 overflow = max VAX double maintaining sign.
	Mantissa shifted piece by piece, lowest 3 bits set to 011 for VAX.
	Output in VAX byte order
*/
#define CVTVAX_DBLE( ibyte, obyte, nbyte )				\
	{ char *last = (ibyte + nbyte);					\
	  while ( ibyte < last )  					\
	   { long  *eee0 = (long *)ibyte;	    /* input double*/	\
	     long  *eee1 = (long *)(ibyte+4);	    /*      pointers*/  \
	     long  vax0,vax1,exp;		    /* work buffers */	\
	     char  *ppp0 = (char *)&vax0;	    /* and pointers */	\
	     char  *ppp1 = (char *)&vax1;				\
	     if (*eee1 == 0 && (*eee0 == 0 && *eee0 == 0x80000000) )	\
	      { vax0 = 0; vax1 = 0; }		     /* VAX zero    */	\
	     else							\
	      { vax0  = (*eee0>>16)&0x00008000;     /* init to sign */	\
		exp   =((*eee0>>20)&0x000007FF)-894;/* exp-1023+129 */	\
	        if (exp < 0x00000000 ) 		    /* underflow ?  */	\
		  vax0 = 0x00000000;		    /* = force zero */  \
		else if (exp > 0xFF ) 	      	    /* overflow ?   */	\
	         { vax0 |= 0xFFFF7FFF; vax1 = 0xFFFFFFFF; }       	\
		else							\
		 { vax0 |= (exp << 7);		    /* load exp     */	\
	           vax0 |= (*eee0>>13)&0x0000007F;  /*  7b at 13->00*/	\
		   vax0 |= (*eee0<<19)&0xFFF80000;  /* 13b at 00->19*/	\
		   vax0 |= (*eee1>>13)&0x00070000;  /*  3b at 29->16*/	\
		   vax1  = (*eee1>>13)&0x0000FFFF;  /* 16b at 13->00*/	\
		   vax1 |= (*eee1<<19)&0xFFF80000;  /* 13b at 00->19*/	\
		   vax1 |= 0x00030000;		    /* load mid.val.*/  \
		 }							\
	      }								\
	     *obyte++ = ppp0[3]; *obyte++ = ppp0[2]; /* output in   */	\
	     *obyte++ = ppp0[1]; *obyte++ = ppp0[0]; /* VAX byte    */	\
	     *obyte++ = ppp0[3]; *obyte++ = ppp0[2]; /* order       */	\
	     *obyte++ = ppp0[1]; *obyte++ = ppp0[0]; ibyte += 8;	\
	   };  								\
	}
#else /*E2M*/
/*
	DOUBLE VAX to 68020 IEEE conversion
	Input in the "VAX" byte order loaded in vax0, vax1
	Output is assumed longword alligned, IEEE is built using pointers
	Exponent changed to a new base (-129+1023) (no overflow danger)
	Mantissa shifted piece by piece, last 3 VAX bits ignored
*/
#define CVTVAX_DBLE( ibyte, obyte, nbyte ) 				\
	{ char *last = (ibyte + nbyte);					\
	  while ( ibyte < last )  					\
	   { long  *eee0 = (long *)obyte;	/* output double */	\
	     long  *eee1 = (long *)(obyte+4);				\
	     long  vax0,vax1;			    /* work buffer  */	\
	     char  *ppp0 = (char *)&vax0;	    /* and pointers */	\
	     char  *ppp1 = (char *)&vax1;				\
	     ppp0[3] = *ibyte++; ppp0[2] = *ibyte++; /* input the   */	\
	     ppp0[1] = *ibyte++; ppp0[0] = *ibyte++; /* VAX ordered */	\
	     ppp1[3] = *ibyte++; ppp1[2] = *ibyte++; /* VAX format  */	\
	     ppp1[1] = *ibyte++; ppp1[0] = *ibyte++;  obyte +=8;	\
	     *eee0  = (vax0<<16)&0x80000000;	    /* init to sign */	\
	     *eee0 |=((vax0<<13)&0x0FF00000)+(894<<20); /*e-129+1023*/  \
	     *eee0 |= (vax0<<13)&0x000FE000;	    /*  7b at 00->13*/	\
	     *eee0 |= (vax0>>19)&0x00001FFF;	    /* 13b at 19->00*/	\
	     *eee1  = (vax0<<13)&0xE0000000;        /*  3b at 16->29*/	\
	     *eee1 |= (vax1<<13)&0x1FFFE000;	    /* 16b at 00->13*/	\
	     *eee1 |= (vax1>>19)&0x00001FFF;	    /* 13b at 19->00*/	\
	   };  								\
	}
#endif
 
#define CVTBYT_DBLE( ibyte, obyte, nbyte ) 	 			\
	{ memcpy (obyte,ibyte,nbyte); }
 
/*#############################################################################
Modified from original by:
Martin Brunecky, Auto-trol Technology Corporation,
12500 North Washington Street, Denver, CO-80241-2404
(303) 252-2499                                        ncar!ico!auto-trol!marbru
#############################################################################*/
